import java.util.*;
import java.io.*;
/**
 * CAPA MODEL
 * Clase Cliente que permite grabar instancias de la misma sobre un archivo relativo <br>
 * que tiene como maximo 1109 registros, no implementa la resolucion de claves sino- <br>
 * nimas, el largo de registro es de 39 bytes, la tecnica de hashing utilizada es<br>
 * utilizar una hashtable que se lee y se graba desde un archivo (que es igual al nombre <br>
 * del archivo de datos con el agregado de ".ndx") de claves. Utiliza RandomAccessFile, todo<br>
 * el acceso al archivo relativo se realiza a traves de la hashtable, a modo de evitar problemas<br>
 * de resolucion de claves.<br>
 * La version 1.0 utilizaba la tecnica del modulo como algoritmo de hashing.<br>
 *<table border=1 align=left>
 *<tr><td><b>tipo de dato</b></td><td><b>campo</b></td><td><b>descripcion</b></td></tr>
 *<tr><td>byte</td><td>ocupado</td><td>0 indica registro vacio, 1 indica registro ocupado</td></tr>
 *<tr><td>int</td><td>codigo</td><td>clave primaria, identificador de cliente entero > 0</td></tr>
 *<tr><td>byte[30]</td><td>razon</td><td>razon social de la cuenta del cliente</td></tr>
 *<tr><td>float</td><td>saldo</td><td>saldo de la cuenta</td></tr>
 *</table>
 * 
 * @autor Guillermo Cherencio
 * @autor UTN - FRD Cisco Networking Academy
 * @version 1.1
 * @see AWTCtrlCliente
 * @see AWTCliente
 * @see CONCliente
 *
*/
public class ClienteHash {
   /** REC_LEN largo de registro utilizado para almacenar instancias */
   public static final int REC_LEN = 39; // 1 byte para ocupado,4 bytes para codigo,30 para razon,4 bytes para saldo
   /** BYTE_PAD byte utilizado para rellenar la razon social hasta 30 caracteres */
   public static final byte BYTE_PAD = (byte) 32; 
   private RandomAccessFile raf;
   private String indice;
   private byte ocupado;
   private int codigo;
   private String razon;
   private byte[] brazon = new byte[30];
   private float saldo;
   private Hashtable hash;
   private boolean esClienteNuevo;
   /** Constructor Cliente 
     * @param archivo String indicando archivo relativo
     * @return nada
   */
   public ClienteHash(String archivo) {
      try {
         raf = new RandomAccessFile(archivo,"rw");
      } catch(IllegalArgumentException iae) {
         System.err.println("Error creando objeto Cliente: " + iae.getMessage());
      } catch(IOException ioe) {
         System.err.println("I/O Error creando objeto Cliente: " + ioe.getMessage());
      } catch(SecurityException se) {
         System.err.println("Falta de Permisos creando objeto Cliente: " + se.getMessage());
      } 
      indice = archivo + ".ndx";
      esClienteNuevo = false;
      hash = new Hashtable();
      leerClaves();
      setCodigo(0);
   }
   /** Constructor Cliente 
     * @param archivo String indicando archivo relativo
     * @param codigo  int    indicando codigo de Cliente
     * @return nada
   */
   public ClienteHash(String archivo,int acodigo) {
      this(archivo);
      setCodigo(acodigo);
   }
   /** Constructor Cliente 
     * @param archivo String indicando archivo relativo
     * @param codigo  int    indicando codigo de Cliente
     * @param razon   String indicando razon social de la cuenta del Cliente
     * @return nada
   */
   public ClienteHash(String archivo,int acodigo,String arazon) {
      this(archivo,acodigo);
      setRazon(arazon);
   }
   /** Constructor Cliente 
     * @param archivo String indicando archivo relativo
     * @param codigo  int    indicando codigo de Cliente
     * @param razon   String indicando razon social de la cuenta del Cliente
     * @param saldo   float  indicando saldo inicial de la cuenta del Cliente
     * @return nada
   */
   public ClienteHash(String archivo,int acodigo,String arazon,float asaldo) {
      this(archivo,acodigo,arazon);
      setSaldo(asaldo);
   }
   /** setOcupado() permite indicar registro ocupado/no ocupado 
     * @param ocupado byte  0 -> vacio, 1 -> ocupado
     * @return nada
   */
   public void setOcupado(byte aocupado) { 
      ocupado = aocupado; 
   }
   /** setCodigo() permite indicar codigo de cliente
     * @param codigo int codigo de cliente > 0
     * @return nada
   */
   public void setCodigo(int acodigo) { 
      codigo = acodigo; 
      esClienteNuevo = existeCliente(acodigo);
      // limpio el contenido del resto de los atributos
      setOcupado((byte) 0);
      setRazon("");
      setSaldo(0F);
   }
   /** setRazon() permite indicar razon social de la cuenta
     * @param razon byte[] razon social, maximo 30 caracteres
     * @return nada
   */
   public void setRazon(byte[] abrazon) {
      int i;
      for(i=0;i<30;i++)
         if (abrazon[i] == 0) abrazon[i] = BYTE_PAD;
      brazon = abrazon;
      razon = new String(brazon);
   }
   /** setRazon() permite indicar razon social de la cuenta
     * @param razon String razon social, maximo 30 caracteres
     * @return nada
   */
   public void setRazon(String arazon) { 
      byte[] blrazon = new byte[30];
      byte[] tmp;
      int i;
      tmp = arazon.getBytes();
      if ( arazon.length() > 30 ) {
         razon = arazon.substring(0,30);
         for(i=0;i<30;i++) blrazon[i] = tmp[i];
      } else {
         razon = arazon; 
         for(i=0;i<arazon.length();i++) blrazon[i] = tmp[i];
         for(;i<30;i++) blrazon[i] = BYTE_PAD;
      }
      brazon = blrazon;
   }
   /** setSaldo() permite indicar saldo inicial de la cuenta
     * @param saldo float saldo inicial de la cuenta
     * @return nada
   */
   public void setSaldo(float asaldo) {
      saldo = asaldo;
   }
   /** getOcupado() devuelve estado del registro
     * @return byte 0 -> vacio, 1 -> ocupado
   */
   public byte getOcupado() { 
      return ocupado; 
   }
   /** getCodigo() devuelve codigo de cliente
     * @return int codigo de cliente
   */
   public int  getCodigo() { 
      return codigo; 
   }
   /** getRazon() devuelve razon social
     * @return String 
   */
   public String getRazon() { 
      return razon; 
   }
   /** getSaldo() devuelve saldo de la cuenta
     * @return float
   */
   public float getSaldo() { 
      return saldo; 
   }
   private byte[] getBRazon() { return brazon; }
   /** cerrar() Cierra el archivo asociado a esta instancia de
       cuenta, requerido antes de dejar de utilizar un objeto
       de esta clase
     * @return nada
   */
   public void cerrar() {
      if ( raf != null ) {
         grabarClaves();
         try {
            raf.close();
         } catch(IOException ioe) {
            System.err.println("I/O Error cerrando archivo relativo Cliente: " + ioe.getMessage());
         } 
      }
      // guardar hash!!!
   }
   /** grabar() almacena los datos de esta instancia
     * @return boolean  true -> Ok, false -> error
   */
   public boolean grabar() {
      posicionarme();
      return grabarCliente();
   }
   private void posicionarme() {
      long dirAbs = dirAbsoluta();
      try {
         raf.seek(dirAbs); //IOe
      } catch(IOException ioe) {
         System.err.println("I/O Error haciendo seek al registro: " + primaryKeyToDirRelativa() + " : " + this + " Error: " + ioe.getMessage());
      }
   }
   /** leer() recupera los datos de esta instancia
     * @return boolean  true -> Ok, false -> error
   */
   public boolean existeCliente(int codigo) {
       Integer i = new Integer(codigo);
       return ( hash.containsKey((Object) i.toString()) );
   }
   /** leer() recupera los datos de esta instancia
     * @return boolean  true -> Ok, false -> error
   */
   public boolean leer() {
      posicionarme();
      return leerCliente();
   }
   /** borrar() borra (libera registro) los datos de esta instancia
     * @return boolean  true -> Ok, false -> error
   */
   public boolean borrar() {
      posicionarme();
      if ( leerCliente() ) {
          setOcupado((byte)0);
          posicionarme();
          if ( grabarCliente() ) {
              borrarClave(getCodigo());
              return true;
          }
      }
      return false;
   }
   /** acreditar() 
     * @param credito float acredita (resta) del saldo de la cuenta
     * @return boolean  true -> Ok, false -> error
   */
   public void acreditar(float acredito) {
      actualizoSaldo((acredito * (-1F)));
   }
   /** debitar() 
     * @param debito float debita (suma) al saldo de la cuenta
     * @return boolean  true -> Ok, false -> error
   */
   public void debitar(float adebito) {
      actualizoSaldo(adebito);
   }
   private void actualizoSaldo(float adebcre) {
      setSaldo(getSaldo() + adebcre);
   }
   /** toString() representacion String de esta instancia
     * Cliente: <codigo> , <razon social> , <saldo>
     * @return String
   */
   public String toString() {  // sobre-escribo metodo toString() de la clase Object
      return "Cliente: " + getCodigo() + "," + getRazon() + "," + getSaldo();
   }
   private boolean grabarCliente() { // graba un cliente en la posicion actual del archivo relativo
      boolean nret = true;
      setOcupado((byte) 1);
      try {
         agregarClave(getCodigo());
         raf.writeByte(getOcupado()); //IOe
         raf.writeInt(getCodigo()); //IOe
         raf.write(getBRazon()); //IOe
         raf.writeFloat(getSaldo()); //IOe
      } catch(IOException ioe) {
         System.err.println("I/O Error grabando " + this + " Error: " + ioe.getMessage());
         nret = false;
      }
      return nret;
   }
   private boolean leerCliente() { // lee un cliente en la posicion actual del archivo relativo
      boolean nret = false;
      byte focupado = (byte) 0;
      int fcodigo = 0, nbytes = (byte) 0;
      byte[] fbrazon = new byte[30];
      float fsaldo = 0F;
      try {
         focupado = raf.readByte(); //EOFe IOe
         if ( focupado == ((byte) 0)) return false;  // registro vacio
         fcodigo = raf.readInt(); //EOFe IOe
         nbytes = raf.read(fbrazon); //IOe -> 30 Ok , -1 EOF
         fsaldo = raf.readFloat(); //EOFe IOe
         // si pude leer el registro completo, actualizo el valor de los atributos
         setOcupado(focupado);
         setCodigo(fcodigo);
         setRazon(fbrazon);
         setSaldo(fsaldo);
         nret = true;
      } catch(EOFException eofe) { // no se pudo recuperar clave grabada, el archivo aun no ha sido ampliado para contener este registro
         System.err.println("EOF Error leyendo " + this + " Error: " + eofe.getMessage());
         System.err.println("EOF Error lei: " + focupado + "," + fcodigo + "," + fbrazon + "," + fsaldo);
         nret = false;
      } catch(IOException ioe) {
         System.err.println("I/O Error leyendo " + this + " Error: " + ioe.getMessage());
         nret = false;
      }
      return nret;
   }
   private int primaryKeyToDirRelativa() { 
      if ( existeCliente(getCodigo()) ) { // registro existente
          return getDir(getCodigo());
      } else { // nuevo registro
          gotoEOF();
          return getCurrentRecord();
      }
   }
   private long dirAbsoluta() { 
      return (long) (primaryKeyToDirRelativa() * REC_LEN); 
   }
   private void gotoEOF() { 
       try {
           raf.seek(raf.length());
       } catch(IOException ioe) {
           System.err.println("Error intentando ir a EOF");
       }
   }
   private int getCurrentRecord() {
       int dir;
       try {
           dir = ((int) (raf.getFilePointer() / REC_LEN));
       } catch(IOException ioe) {
           dir = -1;
       }
       return dir;
   }
   private void agregarClave(int dir,int codigo) {
       Integer valor = new Integer(dir);
       Integer clave = new Integer(codigo);
       hash.put((Object) clave.toString(),(Object) valor);
   }
   private void agregarClave(int codigo) {
       agregarClave(getCurrentRecord(),codigo);
   }
   private void borrarClave(int codigo) {
       Integer clave = new Integer(codigo);
       hash.remove((Object) clave.toString());
   }
   private void leerClaves() { // leo archivo de claves y cargo hashtable
       RandomAccessFile r;
       int clave,valor;
       try {
          r = new RandomAccessFile(this.indice,"r");
          try {
             for ( ; ;) {
                 clave = r.readInt();
                 valor = r.readInt();
                 agregarClave(valor,clave);
             }
          } catch(EOFException eofe) { // EOF
             System.out.println("EOF leyendo claves");
          } catch(IOException ioe) { // IOE
             System.err.println("I/O Error leyendo " + this + " Error: " + ioe.getMessage());
          } finally {
             System.out.println("cierro claves");
             try { r.close(); } catch(IOException ioe) {System.out.println("error cerrando claves");}
          }
       } catch(IllegalArgumentException iae) {
          System.err.println("Error creando archivo indice: " + iae.getMessage());
       } catch(IOException ioe) {
          System.err.println("I/O Error creando archivo indice: " + ioe.getMessage());
       } catch(SecurityException se) {
          System.err.println("Falta de Permisos creando archivo indice: " + se.getMessage());
       }
   }
   private void grabarClaves() { // leo hashtable y grabo archivo claves
       RandomAccessFile r;
       Object co;
       String c;
       Integer cc,v;
       int clave,valor;
       try {
          r = new RandomAccessFile(this.indice,"rw");
          try {
             for (Enumeration e = hash.keys() ; e.hasMoreElements() ;) {
                 co = e.nextElement(); // clave
                 c = (String) co;
                 v = (Integer) hash.get(co);
                 cc = new Integer(c);
                 clave = cc.intValue();
                 valor = v.intValue();
                 r.writeInt(clave);
                 r.writeInt(valor);
             }
          } catch(IllegalArgumentException iae) {
             System.err.println("Error creando archivo indice: " + iae.getMessage());
          } catch(IOException ioe) {
             System.err.println("I/O Error creando archivo indice: " + ioe.getMessage());
          } catch(SecurityException se) {
             System.err.println("Falta de Permisos creando archivo indice: " + se.getMessage());
          } finally {
             System.out.println("cierro grabo claves");
             try { r.close(); } catch(IOException ioe) {System.out.println("error cerrando grabo claves");}
          }
       } catch(IllegalArgumentException iae) {
          System.err.println("Error creando archivo indice: " + iae.getMessage());
       } catch(IOException ioe) {
          System.err.println("I/O Error creando archivo indice: " + ioe.getMessage());
       } catch(SecurityException se) {
          System.err.println("Falta de Permisos creando archivo indice: " + se.getMessage());
       }
   }
   public int getDir(int codigo) {
       Integer i = new Integer(codigo);
       if ( hash.containsKey((Object) i.toString()) ) {
          Integer valor = (Integer) hash.get((Object) i.toString());
          return valor.intValue();
       } else return -1;
   }
   public static void main(String[] args) {
       ClienteHash cl = new ClienteHash("clientes.dat");
       cl.setCodigo(100);
       cl.setRazon("pepe");
       cl.setSaldo(125.35F);
       cl.grabar();
       cl.setCodigo(200);
       cl.setRazon("juancho 200");
       cl.setSaldo(100.55F);
       cl.grabar();
       System.out.println("codigo 100 esta en " + cl.getDir(100));
       System.out.println("codigo 200 esta en " + cl.getDir(200));
       System.out.println("codigo 400 esta en " + cl.getDir(400));
       cl.setCodigo(100);
       cl.leer();
       System.out.println("lei: codigo=" + cl.getCodigo() + " razon=" + cl.getRazon() + " saldo=" + cl.getSaldo());
       cl.setRazon("pepecambio");
       cl.grabar();
       System.out.println("lei: codigo=" + cl.getCodigo() + " razon=" + cl.getRazon() + " saldo=" + cl.getSaldo());
       cl.setCodigo(200);
       cl.leer();
       System.out.println("lei: codigo=" + cl.getCodigo() + " razon=" + cl.getRazon() + " saldo=" + cl.getSaldo());
       cl.setCodigo(400);
       boolean lei = cl.leer();
       System.out.println("lei=" + lei + " codigo=" + cl.getCodigo() + " razon=" + cl.getRazon() + " saldo=" + cl.getSaldo());
       cl.cerrar();
       System.exit(0);
   }
}