import java.util.*;
import java.io.*;
/**
 * Clase HashFile que permite mantener un archivo indexado utilizando una Hashtable + un archivo relativo <br>
 * indicando el largo de registro y el largo de la clave. Tanto la clave
 * como el registro son de tipo String.<br>
 * Clase generica que sirve para cualquier archivo indexado de este tipo, no esta limitado a una estructura de registro<br>
 * determinada.<br>
 * 
 * @autor Guillermo Cherencio
 * @version 1.0
*/
public class HashFile {
    // instance variables - replace the example below with your own
    private String archivo;           // nombre de path - archivo secuencial
    private int reclen;
    private int keylen;
    private RandomAccessFile raf;    
    private String indice;
    private Hashtable hash;
    private String errMsg = null;     // ultimo mensaje de error generado
    private DataOutputStream dout = null;
    private DataInputStream din = null;

    /**
     * Constructor for objects of class HashFile
     */
    public HashFile(final String archivo,final int reclen,final int keylen)
    {
        // initialise instance variables
        setArchivo(archivo);
        setRecLen(reclen);
        setKeyLen(keylen);
        setIndice();
        hash = new Hashtable();
        open();
        leerClaves();
   }
   // GETTERS
    public String getArchivo() { return archivo; }
    public String getErrorMsg() { return errMsg; }
    public String getIndice() { return indice; }
    public int getRecLen() { return reclen; }
    public int getKeyLen() { return keylen; }
   
   // SETTERS
   public void setArchivo(final String archivo) { this.archivo = archivo; }
   public void setRecLen(final int reclen) { this.reclen = reclen; }
   public void setKeyLen(final int keylen) { this.keylen = keylen; }
   private void setIndice() { indice = getArchivo() + ".ndx"; }
   
     /**
     * Establece un mensaje de error
     * @param errMsg String que representa un mensaje de error
     */
   public void setErrorMsg(final String errMsg) { this.errMsg = errMsg;  }
    
   // CUSTOM
   private void leerClaves() {
       File f = new File(getIndice());
       if ( f.exists() ) {
           byte[] bkey = new byte[this.keylen];
           int nreg;
           try {
               din = new DataInputStream(new BufferedInputStream(new FileInputStream(getIndice())));
               for(;;) {
                   din.read(bkey);
                   nreg = din.readInt();
                   String clave = new String(bkey);
                   Integer reg  = new Integer(nreg);
                   hash.put((Object) clave,(Object) reg);
               }
           } catch(EOFException eofe) { // EOF
               System.out.println("EOF leyendo " + getIndice());
           } catch(IOException ioe) {
               errorInFile(ioe,"leer claves " + getIndice());
           } finally {
               try { din.close(); } catch(IOException ioe) { errorInFile(ioe,"cerrar indice " + getIndice()); }
           }
       }
   }
   private void grabarClaves() { // leo hashtable y grabo archivo claves
       Object claveobj;
       Integer reg;
       int clave,valor;
       try {
          dout = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(getIndice())));
          for (Enumeration e = hash.keys() ; e.hasMoreElements() ; ) {
                 claveobj = e.nextElement(); // clave
                 reg = (Integer) hash.get(claveobj);
                 dout.writeBytes((String) claveobj);
                 dout.writeInt(reg.intValue());
             }
       } catch(IOException ioe) {
          errorInFile(ioe,"grabar indice " + getIndice());
       } finally {
          try { dout.close(); } catch(IOException ioe) {errorInFile(ioe,"cerrar indice " + getIndice());}
       }
   }
   private boolean open() {
      try {
         raf = new RandomAccessFile(getArchivo(),"rw");
      } catch(IOException ioe) {
         return errorInFile(ioe,"abrir archivo " + getArchivo());
      }
      return true;
   }
   public void close() {
      if ( raf != null ) {
         grabarClaves();
         try {
            raf.close();
         } catch(IOException ioe) {
            errorInFile(ioe,"cerrar archivo " + getArchivo());
         } 
      }
   }
   private int getDir(final String clave) {
       int bret = -1;
       if ( hash.containsKey((Object) clave) ) {
          bret = ((Integer) hash.get((Object) clave)).intValue();
       } else { 
          try { bret = (int) (raf.length() / this.reclen); /* ultimo registro */ } catch(IOException ioe) { }
       }
       return bret;
   }
   private boolean getPos(final int nreg) {
       try {
           if ( nreg == -1 ) raf.seek(raf.length()); // EOF
           else              raf.seek(nreg * this.reclen);
       } catch(IOException ioe) {
           return errorInFile(ioe,"posicionar registro " + nreg);
       }
       return true;
   }
   public String read(String clave) {
       String bret = null;
       clave = stringPad(clave,this.keylen);
       byte[] reg = new byte[this.reclen];
       int nreg = getDir(clave);
       getPos(nreg);
       try {
         if ( raf.read(reg) != -1) bret = new String(reg);
       } catch(IOException ioe) {
         errorInFile(ioe,"leer registro " + nreg);
       }
       return bret;
   }
   public boolean write(String clave,String registro) {
       clave = stringPad(clave,this.keylen);
       registro = stringPad(registro,this.reclen);
       int nreg = getDir(clave);
       Integer reg = new Integer(nreg);
       getPos(nreg);
       hash.put((Object) clave,(Object) reg);
       try {
         raf.writeBytes(registro); //IOe
       } catch(IOException ioe) {
         return errorInFile(ioe,"grabar registro " + nreg);
       }
       return true;
   }
   public boolean delete(String clave) {
       clave = stringPad(clave,this.keylen);
       return ( hash.remove((Object) clave) != null);
   }
   public boolean reorg() {
       String clave;
       String registro;
       File tmp = null;
       String nuevo_archivo = null;
       try { 
           tmp = File.createTempFile("tmp",".tmp"); 
//System.out.println("tmp1 getabsolutepath:" + tmp.getAbsolutePath());
//System.out.println("tmp2 getcanonicalpath:" + tmp.getCanonicalPath());
//System.out.println("tmp3 getpath:" + tmp.getPath());
           nuevo_archivo = tmp.getPath();
       } catch(IOException ioe) { 
           errorInFile(ioe,"crear archivo temporario"); 
           nuevo_archivo = getArchivo() + "_tmp";
       }
       deleteFile(nuevo_archivo);
       HashFile nuevo = new HashFile(nuevo_archivo,getRecLen(),getKeyLen());
       for (Enumeration e = hash.keys() ; e.hasMoreElements() ; ) {
           clave = (String) e.nextElement();
           registro = read(clave);
           nuevo.write(clave,registro);
       }
       nuevo.close();
       close();
       deleteFile(getArchivo());
       deleteFile(getIndice());
       renameFile(nuevo.getArchivo(),getArchivo());
       renameFile(nuevo.getIndice(),getIndice());
       open();
       return true;
   }
   private boolean deleteFile(final String archivo) {
       File f = new File(archivo);
       if ( f.exists() ) return f.delete();
       return true;
   }
   private boolean renameFile(final String archivo,final String newarchivo) {
       File f1 = new File(archivo);
       File f2 = new File(newarchivo);
       return f1.renameTo(f2);
   }
   private String stringPad(final String txt,final int largo) {
       int largotxt = txt.length();
       if ( largotxt == largo ) return txt;
       if ( largotxt > largo ) return txt.substring(0,largo);
       return txt + space(largo - largotxt);
   }
   private String space(final int n) {
       StringBuffer sb = new StringBuffer("");
       for(int i = 0;i<n;i++) sb.append(" ");
       return sb.toString();
   }
   private boolean errorInFile(IOException ioe,final String op) { 
        setErrorMsg("Error: IOException " + ioe + " intentando " + op + " archivo " + getArchivo()); 
        return false;
   }
     /**
     * Permite saber si se ha producido un error
     * @see clearError()
     * @return Devuelve true en caso de haberse producido algun error, caso contrario,devuelve false
     */
    public boolean hasError() { return ( getErrorMsg() != null );  }
     /**
     * Permite borrar el ultimo mensaje de error generado (si es que habia alguno)
     * @see hasError()
     */
    public void clearError() {  setErrorMsg(null);  }
/////////////////////////////
    public static void main(String[] args) {
       HashFile cl = new HashFile("clientes.dat",30,5); // archivo,largo registro,largo clave
       System.out.println("inicio");
       cl.write("100","Juan");
       cl.write("200","Pedro");
       cl.write("300","Ana");
       cl.write("400","Amalia");
       cl.write("200","Pedrito");
       cl.write("300","Anita");
      for(int i = 100;i<500;i+=100) 
         System.out.println("lei clave " + i + ":" + cl.read("" + i));
      System.out.println("borro clave 200: " + cl.delete("200"));
       System.out.println("grabo clave 200: " + cl.write("200","Nuevo Pedro"));
      for(int i = 100;i<500;i+=100) 
         System.out.println("lei clave " + i + ":" + cl.read("" + i));
       cl.reorg();

/*         
       System.out.println("lei clave 200: " + cl.read("200"));
       System.out.println("lei clave 150: " + cl.read("150"));
       System.out.println("lei clave 300: " + cl.read("300"));
       System.out.println("borro clave 200: " + cl.delete("200"));
       System.out.println("lei clave 200: " + cl.read("200"));
       System.out.println("borro clave 500: " + cl.delete("500"));
       System.out.println("lei clave 500: " + cl.read("500"));
       System.out.println("grabo clave 200: " + cl.write("200","Nuevo Pedro"));
       System.out.println("lei clave 200: " + cl.read("200"));
       cl.reorg();
*/       
       cl.close();
       System.out.println("fin");
   }
///////////////////////////////    
}
